from typing import Iterable, BinaryIO

from elftools.elf.elffile import ELFFile
from elftools.dwarf.die import DIE

from moria.arch import Arch, Endianness
from moria.basetypes import (
    IntType,
    PointerType,
    ArrayType,
    StructField,
    Type,
)
import moria.namespace as ns


DW_ATE_address = 0x1
DW_ATE_boolean = 0x2
DW_ATE_complex_float = 0x3
DW_ATE_float = 0x4
DW_ATE_signed = 0x5
DW_ATE_signed_char = 0x6
DW_ATE_unsigned = 0x7
DW_ATE_unsigned_char = 0x8


class DwarfParser:
    def __init__(self, stream: BinaryIO) -> None:
        self.elf_file = ELFFile(stream)

        if self.elf_file.elfclass in (32, 64):
            word_size = self.elf_file.elfclass // 8
        else:
            raise NotImplementedError(
                f"Unsupported elf class: {self.elf_file.elfclass}"
            )

        endianness = (
            Endianness.LITTLE
            if self.elf_file.little_endian
            else Endianness.BIG
        )

        if not self.elf_file.has_dwarf_info():
            raise ValueError("File has no DWARf debug info.")

        self.dwarf_info = self.elf_file.get_dwarf_info()
        self.namespace = ns.Namespace(Arch(endianness, word_size))

    def create_namespace(self) -> ns.Namespace:
        for compilation_unit in self.dwarf_info.iter_CUs():
            top_DIE = compilation_unit.get_top_DIE()
            self.recurse_die(top_DIE)
        self.namespace.initialize_struct_classes()
        return self.namespace

    def recurse_die(self, die: DIE) -> None:
        if die.tag == "DW_TAG_structure_type":
            self.process_struct_die(die)

        for child in die.iter_children():
            self.recurse_die(child)

    def process_struct_die(self, die: DIE) -> None:
        struct_name = die.attributes["DW_AT_name"].value.decode("utf-8")
        struct = self.namespace.get_or_create_struct_type(struct_name)
        for child in die.iter_children():
            assert child.tag == "DW_TAG_member"

            member_name = child.attributes["DW_AT_name"].value.decode("utf-8")
            member_type = self.resolve_type(child)
            member_offset = child.attributes[
                "DW_AT_data_member_location"
            ].value
            struct.add_field(
                StructField(
                    self.namespace, member_offset, member_type, member_name
                )
            )

    def resolve_type(self, die: DIE) -> Type:
        if (
            die.tag == "DW_TAG_pointer_type"
            and "DW_AT_type" not in die.attributes
        ):
            return IntType(self.namespace, "void", None, False)

        type_die = die.get_DIE_from_attribute("DW_AT_type")
        if type_die.tag == "DW_TAG_pointer_type":
            pointed_type = self.resolve_type(type_die)
            type_size = type_die.attributes.get("DW_AT_byte_size").value
            return PointerType(pointed_type, type_size)
        elif type_die.tag == "DW_TAG_array_type":
            counts = self.get_array_counts(type_die)
            elem_type = self.resolve_type(type_die)
            array_type = elem_type
            for count in counts:
                array_type = ArrayType(self.namespace, array_type, count)
            return array_type
        elif type_die.tag == "DW_TAG_structure_type":
            type_name = type_die.attributes.get("DW_AT_name").value.decode(
                "utf-8"
            )
            return self.namespace.get_or_create_struct_type(type_name)
        elif type_die.tag == "DW_TAG_base_type":
            type_name = type_die.attributes.get("DW_AT_name").value.decode(
                "utf-8"
            )
            type_size = type_die.attributes.get("DW_AT_byte_size").value
            encoding = type_die.attributes.get("DW_AT_encoding").value
            if encoding in (
                DW_ATE_signed,
                DW_ATE_signed_char,
            ):
                return IntType(self.namespace, type_name, type_size, True)
            elif encoding in (
                DW_ATE_boolean,
                DW_ATE_address,
                DW_ATE_unsigned,
                DW_ATE_unsigned_char,
                DW_ATE_float, # TODO proper floating point support
            ):
                return IntType(self.namespace, type_name, type_size, False)
            else:
                raise NotImplementedError(f"Unsupported encoding: 0x{encoding:x}")
        elif type_die.tag == "DW_TAG_typedef":
            if "DW_AT_type" in type_die.attributes:
                return self.resolve_type(type_die)
            type_name = type_die.attributes.get("DW_AT_name").value.decode(
                "utf-8"
            )
            type_size = type_die.attributes.get("DW_AT_byte_size")
            return IntType(
                self.namespace,
                type_name,
                type_size.value if type_size is not None else None,
                False
            )
        else:
            raise NotImplementedError(f"Unknown type tag: {type_die.tag}")

    def get_array_counts(self, die: DIE) -> Iterable[int]:
        counts = []
        for child in die.iter_children():
            if child.tag == "DW_TAG_subrange_type":
                count = child.attributes["DW_AT_upper_bound"].value + 1
                counts.append(count)
        return counts



from __future__ import annotations
from dataclasses import dataclass
from typing import List, Optional


@dataclass
class FreeChunk:
    address: int
    size: int

    @property
    def end(self):
        return self.address + self.size

    def split(self, address: int, size: int) -> List[FreeChunk]:
        end = address + size
        new_chunks: List[FreeChunk] = []
        if (address < self.address
                or end > self.end
                or size < 0):
            raise ValueError('Cannot split chunk {self} by allocating 0x{address:x}+{size}')
        if address > self.address:
            new_chunks.append(FreeChunk(self.address, address - self.address))
        if end < self.end:
            new_chunks.append(FreeChunk(end, self.end - end))
        return new_chunks


class HeapPacker:
    free_chunks: List[FreeChunk]

    def __init__(self, address: int, size: int) -> None:
        self.free_chunks = [FreeChunk(address, size)]

    def alloc(self, size: int, force_address: Optional[int] = None) -> int:
        min_overhead: Optional[int] = None
        best_chunk_idx: Optional[int] = None
        address: Optional[int] = None
        for idx, chunk in enumerate(self.free_chunks):
            if force_address is not None:
                end = force_address + size
                if force_address >= chunk.address and end <= chunk.end:
                    address = force_address
                    best_chunk_idx = idx
                    break
            elif chunk.size >= size:
                overhead = chunk.size - size
                if min_overhead is None or overhead < min_overhead:
                    min_overhead = overhead
                    best_chunk_idx = idx
                    address = chunk.address
        if best_chunk_idx is None:
            raise ValueError('Unable to allocate')
        assert address is not None

        self.free_chunks = (
            self.free_chunks[:best_chunk_idx] +
            self.free_chunks[best_chunk_idx].split(address, size) +
            self.free_chunks[best_chunk_idx+1:]
        )

        return address



import sys

from typing import List

try:
    from moria.parsers.dwarf import DwarfParser
    from moria.util import hexdump
    from moria.values import StructValue
except ImportError:
    # In case moria is not installed, add the parent folder to the module
    # search path so that it's still possible to run this script from the
    # examples directory.
    sys.path.append("../src")
    from moria.parsers.dwarf import DwarfParser
    from moria.util import hexdump
    from moria.values import StructValue

EXAMPLE_BIN_PATH = "cprograms/userlist.bin"


def main():
    try:
        with open(EXAMPLE_BIN_PATH, "rb") as binary:
            ns = DwarfParser(binary).create_namespace()
    except FileNotFoundError:
        print(f"Binary {EXAMPLE_BIN_PATH} not found. Try building it first?")
        return 1

    users: List[StructValue] = [
        ns.user(name="alice"),
        ns.user(name="bob"),
        ns.user(name="charlie"),
    ]

    for i in range(len(users)):
        user = users[i]
        prev_user = users[(i - 1) % len(users)]
        next_user = users[(i + 1) % len(users)]
        user.id = i + 1
        prev_user.next = user.ref()
        user.prev = prev_user.ref()
        next_user.prev = user.ref()
        user.next = next_user.ref()

    start_address = 0x560000000000
    packed = ns.pack_values(start_address, 0x1000, users)
    hexdump(packed, start_address=start_address)

    unpacked_users = ns.array(ns.user, 3).unpack_from_buffer(
        packed, offset=start_address
    )

    for user in unpacked_users.values:
        assert isinstance(user, StructValue)
        print(user)

    return 0


if __name__ == "__main__":
    sys.exit(main())



from typing import Callable, Iterator, Generic, TypeVar, List


def hexdump(data: bytes, start_address: int = 0) -> None:
    chunk_size = 16
    group_size = 8
    for i in range(0, len(data), chunk_size):
        chunk = data[i: i + chunk_size]
        hex_bytes = [f"{byte:02x}" for byte in chunk]
        hex_chunks = []
        for j in range(0, len(chunk), group_size):
            hex_chunk = hex_bytes[j: j + group_size]
            hex_chunks.append(" ".join(hex_chunk))

        data_hex = "  ".join(hex_chunks)
        ascii = "".join([chr(i) if 32 <= i <= 127 else "." for i in chunk])
        print(f"{start_address+i:016x}  {data_hex:<48}  |{ascii}|")


T = TypeVar("T")


class SortedList(Generic[T]):
    def __init__(self, key: Callable[[T, T], int]):
        self.key = key
        self.elements: List[T] = []

    def insert(self, elem: T) -> int:
        min_index = 0
        max_index = len(self.elements)
        while min_index < max_index:
            midpoint = (max_index - min_index) // 2 + min_index
            compare_result = self.key(elem, self.elements[midpoint])
            if compare_result < 0:
                max_index = midpoint
            elif compare_result > 0:
                min_index = midpoint + 1
            else:
                max_index = midpoint
                min_index = midpoint

        assert 0 <= min_index <= len(self.elements)
        self.elements.insert(min_index, elem)
        return min_index

    def __getitem__(self, i: int) -> T:
        return self.elements[i]

    def __len__(self) -> int:
        return len(self.elements)

    def __iter__(self) -> Iterator[T]:
        return (elem for elem in self.elements)



import pytest
from tests.helpers import make_namespace, make_namespaces, amd64_namespace
from moria.basetypes import IntType, PointerType, StructField, StructType


class TestType:
    def test_abstract_methods(self):
        for ns in make_namespaces():
            int_t = ns.Int.type
            with pytest.raises(NotImplementedError):
                super(IntType, int_t).name
            with pytest.raises(NotImplementedError):
                super(IntType, int_t).size
            with pytest.raises(NotImplementedError):
                super(IntType, int_t).__hash__()
            with pytest.raises(NotImplementedError):
                super(IntType, int_t).__eq__(int_t)


class TestIntType:
    def test_unsized_int(self):
        for ns in make_namespaces():
            partial_int_type = IntType(ns, 'partial_int_t', None, False)
            with pytest.raises(ValueError):
                partial_int_type.max

    def test_str_repr(self):
        for ns in make_namespaces():
            assert(str(ns.Int.type) == 'int')
            assert(repr(ns.Int.type) == '<IntType 4 bytes, signed>')

            assert(str(ns.Char.type) == 'char')
            assert(repr(ns.Char.type) == '<IntType 1 byte, signed>')


class TestArrayType:
    def test_name(self):
        for ns in make_namespaces():
            array_t = ns.array(ns.Int, 5).type
            assert array_t.name == 'int[5]'
            assert str(array_t) == 'int[5]'
            assert repr(array_t) == '<ArrayType of 5 int>'

    def test_eq(self):
        for ns in make_namespaces():
            int_arr = ns.array(ns.Int, 5).type
            int_arr2 = ns.array(ns.Int, 5).type
            int8_arr = ns.array(ns.Int8, 5).type
            char_arr = ns.array(ns.Char, 5).type
            char_2_arr = ns.array(ns.Char, 2).type
            assert int_arr == int_arr2
            assert hash(int_arr) == hash(int_arr2)
            assert int8_arr == char_arr
            assert hash(int8_arr) == hash(char_arr)
            assert int_arr != char_arr
            assert char_2_arr != int_arr
            assert char_2_arr != char_arr

    def test_partial(self):
        for ns in make_namespaces():
            partial_int_type = IntType(ns, 'partial_int_t', None, False)
            array_t = ns.array(ns.get_class_for_type(partial_int_type), 5).type
            assert array_t.size is None


class TestPointerType:
    def test_size(self):
        for ns in make_namespaces():
            void_pointer = ns.VoidPointer.type
            assert void_pointer.size == ns.arch.pointer_size

            with pytest.raises(ValueError):
                PointerType(void_pointer, 1)

    def test_hash_eq(self):
        for ns in make_namespaces():
            void_pointer = ns.VoidPointer.type
            assert (void_pointer) == (
                PointerType((IntType(ns, 'void', None, False))))
            assert hash(void_pointer) == hash(
                PointerType((IntType(ns, 'void', None, False))))

    def test_repr(self):
        for ns in make_namespaces():
            void_pointer = ns.VoidPointer.type
            assert repr(void_pointer) == '<PointerType to void>'
            assert repr(void_pointer.get_pointer_type()) == '<PointerType to void*>'
            assert repr(ns.Char.type.get_pointer_type()) == '<PointerType to char>'


class TestStructField:
    def test_repr(self):
        for ns in make_namespaces():
            field = StructField(ns, 0, ns.Int.type, "i")
            assert repr(field) == "<StructField int i>"

    def test_hash_eq(self):
        for ns in make_namespaces():
            field1 = StructField(ns, 0, ns.Int.type, "i")
            field2 = StructField(ns, 0, ns.Int.type, "i")
            field_j = StructField(ns, 0, ns.Int.type, "j")
            field_offset = StructField(ns, 4, ns.Int.type, "i")
            field_char = StructField(ns, 0, ns.Char.type, "i")
            assert hash(field1) == hash(field2)
            assert field1 == field2
            assert field1 != field_j
            assert field1 != field_offset
            assert field1 != field_char


class TestStructType:
    def test_size(self):
        for ns in make_namespaces():
            partial_type = StructType(ns, 'struct partial')
            assert partial_type.size is None

            partial_int_type = IntType(ns, 'partial_int_t', None, False)
            partial_field_struct = StructType(ns, 'struct pfs')
            partial_field_struct.add_field(StructField(ns, 0, partial_int_type, 'partial'))
            assert partial_field_struct.size is None

    def test_eq(self):
        for ns in make_namespaces():
            struct_type = StructType(ns, 'struct test')
            struct_type.add_field(StructField(ns, 0, ns.UInt32.type, 'i'))

            struct_type2 = StructType(ns, 'struct test')
            struct_type2.add_field(StructField(ns, 0, ns.UInt32.type, 'i'))

            assert struct_type is not struct_type2
            assert hash(struct_type) == hash(struct_type2)
            assert struct_type == struct_type2

    def test_reprs(self):
        for ns in make_namespaces():
            child_type = StructType(ns, 'child')
            child_type.add_field(StructField(ns, 0, ns.UInt32.type, 'j'))
            child_type.add_field(StructField(ns, 4, ns.array(ns.Char, 16).type, 'name'))
            assert repr(child_type) == "<StructType child: uint32_t j, char[16] name>"

            struct_type = StructType(ns, 'parent')
            struct_type.add_field(StructField(ns, 0, ns.UInt32.type, 'i'))
            struct_type.add_field(StructField(ns, 4, child_type, 'child'))

            assert struct_type.pretty_string() == '\n'.join([
                'struct parent {',
                '  uint32_t i;',
                '  struct child {',
                '    uint32_t j;',
                '    char[16] name;',
                '  };',
                '};',
            ])



import argparse
import sys
from parsers.dwarf import DwarfParser

EXAMPLE_BIN_PATH = "examples/userlist.bin"


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", type=str, help="Binary file to read")
    args = parser.parse_args()
    try:
        with open(args.input, "rb") as binary:
            namespace = DwarfParser(binary).parse()
    except FileNotFoundError:
        print("File not found.")
        return 1

    namespace.print_structs()
    return 0


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Optional, List
from typing import Type as PyType

from moria.util import SortedList
import moria.namespace as ns
import moria.values as values


class Type(ABC):
    namespace: ns.Namespace

    def __init__(self, namespace: ns.Namespace):
        self.namespace = namespace

    def get_pointer_type(self) -> PointerType:
        return PointerType(referenced_type=self)

    @property
    def value_class(self) -> PyType[values.Value]:
        return self.namespace.get_class_for_type(self)

    @property
    def pointer_class(self) -> PyType[values.PointerValue]:
        pointer_class = self.namespace.get_class_for_type(
            self.get_pointer_type()
        )
        assert issubclass(pointer_class, values.PointerValue)
        return pointer_class

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def size(self) -> Optional[int]:
        raise NotImplementedError()

    @abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError()

    @abstractmethod
    def __eq__(self, other: Type) -> bool:
        raise NotImplementedError()


@dataclass
class IntType(Type):
    namespace: ns.Namespace
    _name: str
    _size: Optional[int]
    _signed: bool

    @property
    def name(self) -> str:
        return self._name

    @property
    def min(self) -> int:
        if self.signed:
            return -self.max - 1
        return 0

    @property
    def max(self) -> int:
        if self.size is None:
            raise ValueError("int of unkown size has no max or min")
        n_bits = self.size * 8
        if self.signed:
            return (1 << ((n_bits) - 1)) - 1
        return (1 << (n_bits)) - 1

    @property
    def signed(self) -> bool:
        return self._signed

    @property
    def size(self) -> Optional[int]:
        return self._size

    def __hash__(self) -> int:
        return hash((self.__class__, self.namespace, self._size, self._signed))

    def __eq__(self, other: Type) -> bool:
        return (
            isinstance(other, IntType)
            and other.namespace is self.namespace
            and other._size == self._size
            and other._signed == self._signed
        )

    def __repr__(self) -> str:
        un = "un" if not self.signed else ""
        s = "s" if self.size != 1 else ""
        return f"<IntType {self.size} byte{s}, {un}signed>"

    def __str__(self) -> str:
        return self.name


@dataclass
class ArrayType(Type):
    namespace: ns.Namespace
    member_type: Type
    count: int

    @property
    def name(self) -> str:
        count_repr = str(self.count) if self.count > 0 else ""
        return f"{self.member_type.name}[{count_repr}]"

    @property
    def size(self) -> Optional[int]:
        if self.member_type.size is not None:
            return self.member_type.size * self.count
        return None

    def __hash__(self) -> int:
        return hash(
            (self.__class__, self.namespace, self.member_type, self.count)
        )

    def __eq__(self, other: Type) -> bool:
        return (
            isinstance(other, ArrayType)
            and other.namespace is self.namespace
            and other.member_type == self.member_type
            and other.count == self.count
        )

    def __repr__(self) -> str:
        return f"<ArrayType of {self.count} {self.member_type}>"

    def __str__(self) -> str:
        return f"{self.member_type}[{self.count}]"


class PointerType(Type):
    referenced_type: Type

    def __init__(self, referenced_type: Type, size: Optional[int] = None):
        super().__init__(referenced_type.namespace)
        self.referenced_type = referenced_type
        if size is not None and size != self.size:
            raise ValueError(
                "Inconsistent pointer sizes! Current namespace "
                f"architecture uses {self.size} bytes, but got {size} bytes."
            )

    @property
    def name(self) -> str:
        return f"{self.referenced_type.name}*"

    @property
    def size(self) -> Optional[int]:
        return self.namespace.arch.pointer_size

    def __hash__(self) -> int:
        return hash((self.__class__, self.namespace, self.referenced_type))

    def __eq__(self, other: Type) -> bool:
        return (
            isinstance(other, PointerType)
            and other.namespace is self.namespace
            and other.referenced_type == self.referenced_type
        )

    def __repr__(self) -> str:
        return f"<PointerType to {self.referenced_type}>"

    def __str__(self) -> str:
        return f"{self.referenced_type}*"


@dataclass
class StructField:
    namespace: ns.Namespace
    offset: int
    field_type: Type
    name: str

    def __repr__(self) -> str:
        return f"<StructField {self.field_type.name} {self.name}>"

    def __str__(self) -> str:
        return f"{self.field_type.name} {self.name}"

    @staticmethod
    def compare_offsets(x: StructField, y: StructField) -> int:
        return x.offset - y.offset

    @property
    def size(self) -> Optional[int]:
        return self.field_type.size

    def __hash__(self) -> int:
        return hash(
            (
                self.__class__,
                self.namespace,
                self.offset,
                self.field_type,
                self.name,
            )
        )

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, StructField)
            and other.namespace is self.namespace
            and self.offset == other.offset
            and self.field_type == other.field_type
            and self.name == other.name
        )


class StructType(Type):
    _name: str
    fields: SortedList[StructField]

    def __init__(self, namespace: ns.Namespace, name: str) -> None:
        super().__init__(namespace)
        self._name = name
        self.fields = SortedList(key=StructField.compare_offsets)

    def add_field(self, field: StructField) -> None:
        self.fields.insert(field)

    def _pretty_lines(self, indent=0) -> List[str]:
        lines = []
        lines.append(f"struct {self.name} {{")
        for field in self.fields:
            if isinstance(field.field_type, StructType):
                lines += field.field_type._pretty_lines(2)
            else:
                lines.append(f"  {field};")
        lines.append(f"}};")
        return [(' '*indent) + line for line in lines]

    def pretty_string(self) -> str:
        return "\n".join(self._pretty_lines())

    def __repr__(self) -> str:
        fields = ", ".join(str(field) for field in self.fields)
        return f"<StructType {self.name}: {fields}>"

    def __str__(self) -> str:
        fields = ", ".join(str(field) for field in self.fields)
        return f"{self.name} {{{fields}}}"

    @property
    def name(self) -> str:
        return self._name

    @property
    def size(self) -> Optional[int]:
        if len(self.fields) == 0:
            return None
        last_field = self.fields[len(self.fields) - 1]
        if last_field.size is None:
            return None
        return last_field.offset + last_field.size

    def __hash__(self) -> int:
        return hash((self.__class__, self.namespace, self.name))

    def __eq__(self, other: Type) -> bool:
        return (
            isinstance(other, StructType)
            and other.namespace is self.namespace
            and self.name == other.name
        )



from __future__ import annotations
from dataclasses import dataclass
import io
import json
from typing import BinaryIO, Iterable, List, Optional, Union

import pytest
from pytest_mock import MockerFixture
from moria.arch import Endianness
from moria.basetypes import IntType, PointerType, StructType

from moria.parsers.dwarf import DW_ATE_address, DW_ATE_signed, DW_ATE_signed_char, DW_ATE_unsigned, DwarfParser
from moria.values import StructValue


@dataclass
class MockDIEAttribute:
    value: Union[int, bytes]
    die_ref_name: Optional[str] = None
    die_ref: Optional[MockDIE] = None


class MockDIE:
    name: Optional[str]
    tag: str
    attributes: dict[str, MockDIEAttribute]
    children: List[MockDIE]

    def __init__(self,
                 tag: str,
                 attributes: dict[str, MockDIEAttribute],
                 children: Iterable[MockDIE],
                 name: Optional[str] = None) -> None:
        self.tag = tag
        self.attributes = attributes
        self.children = list(children)
        self.name = name

    def iter_children(self) -> Iterable[MockDIE]:
        return self.children

    def get_DIE_from_attribute(self, attribute_name: str) -> MockDIE:
        die = self.attributes[attribute_name].die_ref
        assert die is not None
        return die


class MockCompilationUnit:
    top_die: MockDIE
    dies_by_name: dict[str, MockDIE]
    all_dies: List[MockDIE]

    def __init__(self, json_data: dict) -> None:
        die_data = json_data['die']
        self.all_dies = []
        self.dies_by_name = {}
        self.top_die = self.parse_die(die_data)
        self.resolve_die_refs()

    def resolve_die_refs(self):
        for die in self.all_dies:
            for attribute in die.attributes.values():
                if attribute.die_ref_name is not None:
                    attribute.die_ref = self.dies_by_name[attribute.die_ref_name]

    def parse_die(self, die_data: dict) -> MockDIE:
        children_data = die_data.get('children')
        children = []
        if children_data is not None:
            children = [self.parse_die(child_data) for child_data in children_data]

        attributes = {}
        attributes_data = die_data.get('attributes')
        if attributes_data is not None:
            for attribute_data in attributes_data:
                name = attribute_data['name']
                value = attribute_data.get('value')
                if isinstance(value, str):
                    value = value.encode('utf=8')
                attributes[name] = MockDIEAttribute(
                    value, die_ref_name=attribute_data.get('die_ref'))

        die_name = die_data.get('name')
        die = MockDIE(die_data['tag'], attributes, children, die_name)
        self.all_dies.append(die)
        if die_name is not None:
            self.dies_by_name[die_name] = die
        return die

    def get_top_DIE(self) -> MockDIE:
        return self.top_die


class MockDwarfInfo:
    compilation_units: List[MockCompilationUnit]

    def __init__(self, json_data: dict) -> None:
        self.compilation_units = [MockCompilationUnit(
            cu_data) for cu_data in json_data['compilation_units']]

    def iter_CUs(self) -> Iterable[MockCompilationUnit]:
        return self.compilation_units


class MockELFFile:
    elfclass: int
    little_endian: bool
    dwarf_info: Optional[MockDwarfInfo]

    def __init__(self, stream: BinaryIO) -> None:
        json_data = json.load(stream)
        self.elfclass = json_data['elfclass']
        self.little_endian = json_data['little_endian']
        dwarf_info_data = json_data.get('dwarf_info')
        self.dwarf_info = MockDwarfInfo(
            dwarf_info_data) if dwarf_info_data is not None else None

    def has_dwarf_info(self) -> bool:
        return self.dwarf_info is not None

    def get_dwarf_info(self) -> MockDwarfInfo:
        assert self.dwarf_info is not None
        return self.dwarf_info


int_die = {
    'name': 'int',
    'tag': 'DW_TAG_base_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'int',
        },
        {
            'name': 'DW_AT_byte_size',
            'value': 4,
        },
        {
            'name': 'DW_AT_encoding',
            'value': DW_ATE_signed,
        },
    ],
}

void_ptr_die = {
    'name': 'void',
    'tag': 'DW_TAG_pointer_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'void',
        },
        {
            'name': 'DW_AT_byte_size',
            'value': 8,
        },
        {
            'name': 'DW_AT_encoding',
            'value': DW_ATE_address,
        },
    ],
}

char_die = {
    'name': 'char',
    'tag': 'DW_TAG_base_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'char',
        },
        {
            'name': 'DW_AT_byte_size',
            'value': 1,
        },
        {
            'name': 'DW_AT_encoding',
            'value': DW_ATE_signed_char,
        },
    ],
}

uint64_die = {
    'name': 'uint64_t',
    'tag': 'DW_TAG_base_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'uint64_t',
        },
        {
            'name': 'DW_AT_byte_size',
            'value': 8,
        },
        {
            'name': 'DW_AT_encoding',
            'value': DW_ATE_unsigned,
        },
    ],
}

typedef_int_num_die = {
    'name': 'num',
    'tag': 'DW_TAG_typedef',
    'attributes': [
        {
            'name': 'DW_AT_type',
            'die_ref': 'int',
        },
    ],
}

typedef_partial_die = {
    'name': 'partial_t',
    'tag': 'DW_TAG_typedef',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'partial_t',
        },
    ],
}

partial_ptr_die = {
    'name': 'partial_ptr',
    'tag': 'DW_TAG_pointer_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'void',
        },
        {
            'name': 'DW_AT_type',
            'die_ref': 'partial_t',
        },
        {
            'name': 'DW_AT_byte_size',
            'value': 8,
        },
        {
            'name': 'DW_AT_encoding',
            'value': DW_ATE_address,
        },
    ],

}

char_array_die = {
    'name': 'char_array',
    'tag': 'DW_TAG_array_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'char[16]',
        },
        {
            'name': 'DW_AT_type',
            'die_ref': 'char',
        },
        {
            'name': 'DW_AT_encoding',
            'value': DW_ATE_address,
        },
    ],
    'children': [
        {
            'tag': 'DW_TAG_subrange_type',
            'attributes': [
                {
                    'name': 'DW_AT_upper_bound',
                    'value': 15,
                },
            ]
        }
    ],
}

child_struct_die = {
    'name': 'child_struct',
    'tag': 'DW_TAG_structure_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'child_struct',
        },
    ],
    'children': [
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'child_uint64_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'uint64_t',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 0,
                },
            ],
        },
    ],
}

struct_die = {
    'tag': 'DW_TAG_structure_type',
    'attributes': [
        {
            'name': 'DW_AT_name',
            'value': 'test',
        },
    ],
    'children': [
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'int_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'int',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 0,
                },
            ],
        },
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'void_ptr_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'void',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 8,
                },
            ],
        },
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'char_arr_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'char_array',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 16,
                },
            ],
        },
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'nested_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'child_struct',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 32,
                },
            ],
        },
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'num_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'num',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 40,
                },
            ],
        },
        {
            'tag': 'DW_TAG_member',
            'attributes': [
                {
                    'name': 'DW_AT_name',
                    'value': 'partial_ptr_field',
                },
                {
                    'name': 'DW_AT_type',
                    'die_ref': 'partial_ptr',
                },
                {
                    'name': 'DW_AT_data_member_location',
                    'value': 44,
                },
            ],
        },
    ],
}

elf_data_amd64 = {
    'elfclass': 64,
    'little_endian': True,
    'dwarf_info': {
        'compilation_units': [
            {
                'die': {
                    'tag': 'DW_TAG_compile_unit',
                    'children': [
                        typedef_int_num_die,
                        typedef_partial_die,
                        partial_ptr_die,
                        uint64_die,
                        char_die,
                        char_array_die,
                        int_die,
                        struct_die,
                        void_ptr_die,
                        child_struct_die,
                    ]
                }
            }
        ]
    }
}


class TestDwarf:
    def test_unsupported_tag(self, mocker: MockerFixture) -> None:
        mocker.patch('moria.parsers.dwarf.ELFFile', new=MockELFFile)
        deep_elf_copy = json.loads(json.dumps(elf_data_amd64))
        uint64_die = deep_elf_copy['dwarf_info']['compilation_units'][0]['die']['children'][3]
        uint64_die['tag'] = 'asdf'

        stream = io.BytesIO(json.dumps(deep_elf_copy).encode('utf-8'))
        with pytest.raises(NotImplementedError):
            DwarfParser(stream).create_namespace()

    def test_unsupported_encoding(self, mocker: MockerFixture) -> None:
        mocker.patch('moria.parsers.dwarf.ELFFile', new=MockELFFile)
        deep_elf_copy = json.loads(json.dumps(elf_data_amd64))
        uint64_die = deep_elf_copy['dwarf_info']['compilation_units'][0]['die']['children'][3]
        uint64_die['attributes'][2]['value'] = 0x1337

        stream = io.BytesIO(json.dumps(deep_elf_copy).encode('utf-8'))
        with pytest.raises(NotImplementedError):
            DwarfParser(stream).create_namespace()

    def test_unsupported_class(self, mocker: MockerFixture) -> None:
        mocker.patch('moria.parsers.dwarf.ELFFile', new=MockELFFile)
        bad_elf = {
            'elfclass': 63,
            'little_endian': True,
            'dwarf_info': {
                'compilation_units': []
            }
        }
        stream = io.BytesIO(json.dumps(bad_elf).encode('utf-8'))
        with pytest.raises(NotImplementedError):
            DwarfParser(stream).create_namespace()

    def test_no_dwarf(self, mocker: MockerFixture) -> None:
        mocker.patch('moria.parsers.dwarf.ELFFile', new=MockELFFile)
        bad_elf = {
            'elfclass': 64,
            'little_endian': True,
        }
        stream = io.BytesIO(json.dumps(bad_elf).encode('utf-8'))
        with pytest.raises(ValueError):
            DwarfParser(stream).create_namespace()

    def test_dwarf(self, mocker: MockerFixture, capfd: pytest.CaptureFixture) -> None:
        mocker.patch('moria.parsers.dwarf.ELFFile', new=MockELFFile)
        stream = io.BytesIO(json.dumps(elf_data_amd64).encode('utf-8'))
        ns = DwarfParser(stream).create_namespace()
        assert ns.arch.pointer_size == 8
        assert ns.arch.endianness == Endianness.LITTLE
        assert issubclass(ns.test, StructValue)
        assert isinstance(ns.test.type, StructType)
        assert len(ns.test.type.fields) == 6

        assert ns.test.type.fields[0].name == 'int_field'
        assert ns.test.type.fields[0].field_type == ns.Int.type

        assert ns.test.type.fields[1].name == 'void_ptr_field'
        assert ns.test.type.fields[1].field_type == ns.VoidPointer.type

        assert ns.test.type.fields[2].name == 'char_arr_field'
        assert ns.test.type.fields[2].field_type == ns.array(ns.Char, 16).type

        assert ns.test.type.fields[3].name == 'nested_field'
        assert ns.test.type.fields[3].field_type == ns.child_struct.type

        assert ns.test.type.fields[4].name == 'num_field'
        assert isinstance(ns.test.type.fields[4].field_type, IntType)
        assert ns.test.type.fields[4].field_type.signed == True
        assert ns.test.type.fields[4].field_type.size == 4

        assert ns.test.type.fields[5].name == 'partial_ptr_field'
        assert isinstance(ns.test.type.fields[5].field_type, PointerType)
        assert isinstance(ns.test.type.fields[5].field_type.referenced_type, IntType)
        assert ns.test.type.fields[5].field_type.referenced_type.size is None

        ns.print_structs()
        out, err = capfd.readouterr()
        assert err == ''
        assert out == '''
struct test {
  int int_field;
  void* void_ptr_field;
  char[16] char_arr_field;
  struct child_struct {
    uint64_t child_uint64_field;
  };
  int num_field;
  partial_t* partial_ptr_field;
};

struct child_struct {
  uint64_t child_uint64_field;
};
'''[1:]

    def test_invalid(self) -> None:
        stream = io.BytesIO(b"definitely not a valid ELF file")
        with pytest.raises(Exception):
            DwarfParser(stream)



from moria.pack import FreeChunk, HeapPacker
import pytest

class TestPacker:
    def test_bad_split(self) -> None:
        with pytest.raises(ValueError):
            FreeChunk(0, 1).split(0, 2)

    def test_pack(self) -> None:
        packer = HeapPacker(0, 3)
        assert packer.alloc(1, force_address=1) == 1
        assert len(packer.free_chunks) == 2
        with pytest.raises(ValueError):
            packer.alloc(1, force_address=1)
        with pytest.raises(ValueError):
            packer.alloc(2) 

        addr1 = packer.alloc(1) 
        assert len(packer.free_chunks) == 1
        addr2 = packer.alloc(1)
        assert sorted([addr1, addr2]) == [0, 2]
        assert len(packer.free_chunks) == 0
        with pytest.raises(ValueError):
            packer.alloc(1) 




from __future__ import annotations
from typing import (
    ClassVar,
    Iterable,
    Mapping,
    MutableMapping,
    List,
    Optional,
    Sequence,
    Union,
)

from abc import ABC, abstractmethod
from typing import Type as PyType

from moria.arch import Endianness
import moria.namespace as ns
import moria.basetypes as basetypes


def _pack_integral_type(
    num: int, size: int, signed: bool, endianness: Endianness
) -> bytes:
    n_bits = size * 8
    if signed:
        max_value = 1 << (n_bits) - 1
        min_value = -max_value - 1
    else:
        min_value = 0
        max_value = (1 << (n_bits + 1)) - 1

    un = "un" if not signed else ""
    if not min_value <= num <= max_value:
        raise ValueError(
            f"Cannot represent number {num} in "
            f"{un}signed integral type of size {size} bytes"
        )

    mask = sum(0xFF << (i * 8) for i in range(size))
    num &= mask
    little_endian_bytes = [(num >> (i * 8)) & 0xFF for i in range(size)]
    if endianness == Endianness.BIG:
        return bytes(little_endian_bytes[::-1])
    return bytes(little_endian_bytes)


def _unpack_integral_type(
    buffer: bytes, size: int, signed: bool, endianness: Endianness
) -> int:
    if endianness == endianness.BIG:
        buffer = buffer[::-1]
    num = 0
    assert len(buffer) == size
    for i in range(size):
        num |= buffer[i] << (8 * i)
    sign_bit = 0x80 << (8 * (size - 1))
    if signed and (num & sign_bit):
        mask = sum(0xFF << (i * 8) for i in range(size))
        return -((~num + 1) & mask)
    return num


class Value(ABC):
    # Python types that can potentially be converted to a Value
    CompatibleType = Union[
        "Value", str, int, bytes, float, Iterable["CompatibleType"]
    ]

    # Class variable that will be assigned in subtypes
    type: ClassVar[basetypes.Type]

    address_base: Optional[Value]
    offset: Optional[int]

    def __init__(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> None:
        self.address_base: Optional[Value] = address_base
        self.offset: Optional[int] = offset

    @classmethod
    def get_pointer_class(
        cls,
    ) -> PyType[PointerValue]:
        pointer_class = cls.type.namespace.get_class_for_type(cls.type.get_pointer_type())
        assert issubclass(pointer_class, PointerValue)
        return pointer_class

    @classmethod
    def unpack_from_buffer(
        cls,
        buffer: bytes,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ):
        raise NotImplementedError()

    @classmethod
    def cast(
        cls,
        value: CompatibleType,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> Value:
        raise NotImplementedError()

    def move(
        self,
        address_base: Optional[Value],
        offset: Optional[int],
    ) -> None:
        self.address_base = address_base
        self.offset = offset

    @property
    def address(self) -> Optional[int]:
        """Compute the address of the first byte of the value.
        This may be None if the value has not yet been resolved to an address.
        For values with a base_address value, the address is computed relative
        to this value. Otherwise, offset is used as an absolute address."""
        addr = self.offset
        if addr is not None and self.address_base is not None:
            if self.address_base.address is not None:
                return self.address_base.address + addr
            return None
        return addr

    @property
    def namespace(self) -> ns.Namespace:
        return self.type.namespace

    @abstractmethod
    def copy(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> Value:
        raise NotImplementedError()

    def ref(self) -> PointerValue:
        return self.type.pointer_class(referenced_value=self)

    @abstractmethod
    def pack(self) -> bytes:
        raise NotImplementedError()

    @abstractmethod
    def iter_referenced_values(self) -> Iterable[Value]:
        raise NotImplementedError()


class IntValue(Value):
    type: ClassVar[basetypes.IntType]
    value: Optional[int]

    def __init__(
        self,
        value: Optional[int] = None,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ):
        super().__init__(address_base, offset)
        if value and not self.type.min <= value <= self.type.max:
            un = "un" if not self.type.signed else ""
            raise ValueError(
                f"Integer {value} cannot be "
                f"represented by {un}signed int of size {self.type.size}"
            )
        self.value = value

    @classmethod
    def unpack_from_buffer(
        cls,
        buffer: bytes,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ):
        assert cls.type.size is not None
        if len(buffer) != cls.type.size:
            raise ValueError(
                f"Invalid length to unpack type {cls.type}: "
                f"need {cls.type.size} bytes, got {len(buffer)}"
            )
        val = _unpack_integral_type(
            buffer,
            cls.type.size,
            cls.type.signed,
            cls.type.namespace.arch.endianness,
        )
        return cls(value=val, address_base=address_base, offset=offset)

    @classmethod
    def cast(
        cls,
        value: Value.CompatibleType,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> IntValue:
        if cls.type.size is None:
            raise ValueError("Cannot cast to integer type without a size")

        int_val: Optional[int] = None
        if isinstance(value, int):
            int_val = value
        elif isinstance(value, float):
            int_val = int(value)
        elif isinstance(value, str) and len(value) == 1:
            int_val = ord(value)
        elif isinstance(value, bytes) and len(value) == 1:
            int_val = value[0]

        if int_val is not None:
            mask = sum(0xFF << (i * 8) for i in range(cls.type.size))
            num = int_val & mask
            if num > cls.type.max:
                num = -((~num + 1) & mask)

            return cls(
                value=num,
                address_base=address_base,
                offset=offset,
            )

        raise TypeError(
            f"Type {cls.type} cannot be assigned from {type(value)}"
        )

    def iter_referenced_values(self) -> Iterable[Value]:
        if self.address_base is not None:
            yield self.address_base

    def copy(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> IntValue:
        return self.__class__(
            value=self.value,
            address_base=address_base,
            offset=offset,
        )

    def __float__(self) -> float:
        return float(int(self))

    def __int__(self) -> int:
        if self.value is None:
            raise ValueError(
                "Cannot convert uninitialized integer value to int"
            )
        return self.value

    def __repr__(self) -> str:
        value_str = (
            str(self.value) if self.value is not None else "<uninitialized>"
        )
        return f"<{self.type.name} {value_str}>"

    def __str__(self) -> str:
        return str(self.value) if self.value is not None else "<uninitialized>"

    def pack(self) -> bytes:
        assert (
            self.type.size is not None
        ), "Cannot serialize type of unresolved size"
        val = self.value
        if val is None:
            val = 0
        return _pack_integral_type(
            val,
            self.type.size,
            signed=self.type.signed,
            endianness=self.namespace.arch.endianness,
        )


class ArrayValue(Value):
    type: basetypes.ArrayType

    def __init__(
        self,
        values: Optional[List[Value]] = None,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ):
        super().__init__(address_base, offset)

        assert self.type.member_type.size is not None
        if values is None:
            self.values: List[Value] = [
                self.type.member_type.value_class(
                    address_base=self,
                    offset=i * self.type.member_type.size,
                )
                for i in range(self.type.count)
            ]
        else:
            if len(values) != self.type.count:
                raise ValueError(
                    "Wrong number of values to initialize array. "
                    f"Expected {self.type.count} items, got {len(values)}."
                )
            self.values = [
                val.copy(
                    address_base=self, offset=i * self.type.member_type.size
                )
                for i, val in enumerate(values)
            ]

    @classmethod
    def unpack_from_buffer(
        cls,
        buffer: bytes,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ):
        assert cls.type.member_type.size is not None
        count = cls.type.count
        item_size = cls.type.member_type.size
        assert len(buffer) == count * item_size

        member_class = cls.type.member_type.value_class
        vals = []
        item_offset = 0
        for i in range(count):
            vals.append(
                member_class.unpack_from_buffer(
                    buffer[item_offset: item_offset + item_size],
                )
            )
            item_offset += item_size
        return cls(vals, address_base=address_base, offset=offset)

    def iter_referenced_values(self) -> Iterable[Value]:
        if self.address_base is not None:
            yield self.address_base
        yield from self.values

    @classmethod
    def cast(
        cls,
        value: Value.CompatibleType,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> ArrayValue:
        if not isinstance(value, Iterable):
            raise TypeError(
                f"Type {cls.type} cannot be assigned from {type(value)}"
            )
        value_list = list(value)
        if len(value_list) > cls.type.count:
            raise TypeError(
                f"Too many elements ({len(value_list)}) to fit in {cls.type}"
            )

        values: List[Value] = []
        member_class = cls.type.member_type.value_class
        for item in value_list:
            values.append(member_class.cast(item))

        padding = cls.type.count - len(value_list)
        for _ in range(padding):
            val = member_class()
            values.append(val)

        return cls(values, address_base, offset)

    def copy(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> ArrayValue:
        return self.__class__(
            values=self.values,
            address_base=address_base,
            offset=offset,
        )

    def is_character_array(self) -> bool:
        return (
            isinstance(self.type.member_type, basetypes.IntType)
            and self.type.member_type.size == 1
        )

    def __bytes__(self) -> bytes:
        if not self.is_character_array():
            raise TypeError(
                "Byte conversion is only supported for int8_t and uint8_t arrays."
            )

        char_vals = []
        for val in self.values:
            assert isinstance(val, IntValue) and val.type.size == 1
            if val.value is None or val.value == 0:
                break
            char_vals.append(val.value)
        return bytes(char_vals)

    def __len__(self) -> int:
        return self.type.count

    def __str__(self) -> str:
        return bytes(self).decode("utf-8")

    def __repr__(self) -> str:
        if self.is_character_array():
            return f"(char[{self.type.count}]){bytes(self)!r}"
        else:
            return "{" + ", ".join(str(val) for val in self.values) + "}"

    def pack(self) -> bytes:
        parts: List[bytes] = []

        assert (
            self.type.member_type.size is not None
        ), "Cannot serialize array elements of unresolved size"

        for elem in self.values:
            part = elem.pack()
            assert len(part) == self.type.member_type.size
            parts.append(part)

        return b"".join(parts)


class PointerValue(Value):
    type: ClassVar[basetypes.PointerType]
    raw_value: Optional[int]

    def __init__(
        self,
        referenced_value: Optional[Value] = None,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
        raw_value: int = None,
    ):
        super().__init__(address_base, offset)
        if referenced_value is not None and raw_value is not None:
            raise ValueError(
                "Cannot have both a referenced object and a raw value."
            )
        self.referenced_value: Optional[Value] = referenced_value
        self.raw_value = raw_value

    @classmethod
    def unpack_from_buffer(
        cls,
        buffer: bytes,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ):
        val = _unpack_integral_type(
            buffer,
            cls.type.namespace.arch.pointer_size,
            False,
            cls.type.namespace.arch.endianness,
        )
        return cls(address_base=address_base, offset=offset, raw_value=val)

    @classmethod
    def cast(
        cls,
        value: Value.CompatibleType,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> PointerValue:
        if isinstance(value, str):
            value = value.encode("utf-8")

        if isinstance(value, Sequence):
            array_class = cls.type.namespace.array(
                cls.type.referenced_type.value_class, len(value))
            buffer = array_class(
                [
                    cls.type.referenced_type.value_class.cast(v).copy()
                    for v in value
                ],
            )
            return cls(
                referenced_value=buffer,
                address_base=address_base,
                offset=offset,
            )
        elif isinstance(value, int):
            return cls(
                address_base=address_base,
                offset=offset,
                raw_value=value,
            )
        raise TypeError(
            f"Type {cls.type} cannot be assigned from {type(value)}"
        )

    def iter_referenced_values(self) -> Iterable[Value]:
        if self.address_base is not None:
            yield self.address_base
        if self.referenced_value is not None:
            yield self.referenced_value

    def copy(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> PointerValue:
        return self.__class__(
            referenced_value=self.referenced_value,
            address_base=address_base,
            offset=offset,
            raw_value=self.raw_value,
        )

    @property
    def pointed_address(self) -> Optional[int]:
        if self.raw_value is not None:
            return self.raw_value
        if self.referenced_value is None:
            return None
        if self.referenced_value.address is None:
            return None
        return self.referenced_value.address

    def __repr__(self) -> str:
        addr = self.pointed_address
        return f"<PointerValue to {self.type.referenced_type}@{addr}>"

    def __int__(self) -> int:
        if self.pointed_address is None:
            raise ValueError("Pointer has no initialized value.")
        return self.pointed_address

    def __str__(self) -> str:
        pointed_type = (
            self.referenced_value.type.name
            if self.referenced_value is not None
            else "void"
        )
        if self.pointed_address is None:
            addr = "None"
        elif self.pointed_address == 0:
            addr = f"NULL"
        else:
            addr = f"0x{self.pointed_address:x}"
        return f"({pointed_type}*){addr}"

    def pack(self) -> bytes:
        assert (
            self.type.size is not None
        ), "Cannot serialize pointer of unresolved size"

        pointed_address = self.pointed_address
        if pointed_address is None:
            pointed_address = 0

        return _pack_integral_type(
            pointed_address,
            self.type.size,
            signed=False,
            endianness=self.namespace.arch.endianness,
        )


class StructValue(Value):
    type: ClassVar[basetypes.StructType]
    fields: MutableMapping[str, Value]
    field_types: Mapping[str, basetypes.Type]

    def __init__(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
        **kwargs: Value.CompatibleType,
    ):
        self._initialized: bool = False
        super().__init__(address_base, offset)
        self.fields = {
            field.name: field.field_type.value_class(
                address_base=self, offset=field.offset
            )
            for field in self.type.fields
        }
        self.field_types = {
            field.name: field.field_type for field in self.type.fields
        }
        self._initialized = True
        for field_name, value in kwargs.items():
            setattr(self, field_name, value)

    @classmethod
    def unpack_from_buffer(
        cls,
        buffer: bytes,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> StructValue:
        assert len(buffer) == cls.type.size
        values = {}
        for field in cls.type.fields:
            assert field.size is not None
            buffer_part = buffer[field.offset: field.offset + field.size]
            values[
                field.name
            ] = field.field_type.value_class.unpack_from_buffer(buffer_part)
        return cls(address_base=address_base, offset=offset, **values)

    def iter_referenced_values(self) -> Iterable[Value]:
        if self.address_base is not None:
            yield self.address_base
        yield from self.fields.values()

    @classmethod
    def cast(
        cls,
        value: Value.CompatibleType,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> StructValue:
        if isinstance(value, StructValue) and value.type == cls.type:
            return value
        raise TypeError(
            f"Type {cls.type} cannot be cast from {type(value)}"
        )

    def copy(
        self,
        address_base: Optional[Value] = None,
        offset: Optional[int] = None,
    ) -> StructValue:
        struct = self.__class__(address_base=address_base, offset=offset)
        for name, field in self.fields.items():
            setattr(struct, name, field)
        return struct

    def __repr__(self):
        fields = " ".join(
            f"{field.name}={self.fields[field.name]}"
            for field in self.type.fields
        )
        addr = f"0x{self.address:x}" if self.address is not None else "None"
        return f"<struct {self.type.name} @{addr}: {fields}>"

    def __getattr__(self, name: str) -> Value:
        attr = self.fields.get(name)
        if attr is not None:
            return attr
        raise AttributeError(f"Struct {self.type.name} has no field {name}")

    def __setattr__(self, name: str, val: Value.CompatibleType) -> None:
        if name == "_initialized" or not self._initialized:
            return super().__setattr__(name, val)

        if name not in self.fields:
            return super().__setattr__(name, val)
        field_type = self.fields[name].type
        offset = self.fields[name].offset

        if isinstance(val, field_type.value_class):
            wrappedVal = val.copy(address_base=self, offset=offset)
        else:
            wrappedVal = field_type.value_class.cast(
                val, address_base=self, offset=offset
            )
        self.fields[name] = wrappedVal

    def pack(self) -> bytes:
        parts: List[bytes] = []
        last_field_end = 0
        for type_field in self.type.fields:
            padding = type_field.offset - last_field_end
            assert padding >= 0

            field_value = self.fields[type_field.name]
            try:
                field_bytes = field_value.pack()
            except Exception as err:
                raise Exception(f"unable to pack field {type_field.name}", err)

            assert (
                type_field.size is not None
            ), "Cannot serialize field of unresolved size"
            assert len(field_bytes) == type_field.size

            parts.append(b"\x00" * padding + field_bytes)
            last_field_end = type_field.offset + type_field.size

        return b"".join(parts)



import pytest

from moria.arch import Endianness
from moria.basetypes import IntType, StructField
from moria.values import (
    ArrayValue,
    IntValue,
    PointerValue,
    StructValue,
    Value,
    _pack_integral_type,
)
from tests.helpers import make_namespace, make_namespaces, amd64_namespace


class TestStructValue:
    def test_pack(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        struct_type = ns.get_or_create_struct_type('test')
        struct_type.add_field(StructField(ns, 0, ns.Int.type, 'int_field'))
        struct_type.add_field(StructField(
            ns, 4, struct_type.get_pointer_type(), 'self_ptr'))

        struct_class = ns.get_class_for_type(struct_type)
        assert issubclass(struct_class, StructValue)

        struct = struct_class(int_field=1, self_ptr=2)
        assert struct.pack() == b'\x01\x00\x00\x00' b'\x02\x00\x00\x00'

        unpackable_struct_type = ns.get_or_create_struct_type('unpackable')
        unpackable_struct_type.add_field(StructField(ns, 0, IntType(
            ns, 'partial int', None, False), 'unpackable_field'))
        unpackable_struct_class = ns.get_class_for_type(unpackable_struct_type)
        assert issubclass(unpackable_struct_class, StructValue)
        with pytest.raises(Exception, match='.*unpackable_field.*'):
            unpackable_struct_class().pack()

    def test_repr(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        struct_type = ns.get_or_create_struct_type('test')
        struct_type.add_field(StructField(ns, 0, ns.Int.type, 'int_field'))
        struct_type.add_field(StructField(
            ns, 4, struct_type.get_pointer_type(), 'self_ptr'))

        struct_class = ns.get_class_for_type(struct_type)
        assert issubclass(struct_class, StructValue)

        struct = struct_class(int_field=1, self_ptr=2)

        assert repr(struct) == '<struct test @None: int_field=1 self_ptr=(void*)0x2>'

        struct.move(address_base=None, offset=0)
        assert repr(struct) == '<struct test @0x0: int_field=1 self_ptr=(void*)0x2>'

    def test_cast(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        struct_type = ns.get_or_create_struct_type('test')
        struct_type.add_field(StructField(ns, 0, ns.Int.type, 'int_field'))
        struct_type.add_field(StructField(
            ns, 4, struct_type.get_pointer_type(), 'self_ptr'))

        struct_class = ns.get_class_for_type(struct_type)
        assert issubclass(struct_class, StructValue)

        struct = struct_class.cast(struct_class(int_field=1, self_ptr=2))
        assert isinstance(struct.int_field, IntValue)
        assert int(struct.int_field) == 1
        assert isinstance(struct.self_ptr, PointerValue)
        assert int(struct.self_ptr) == 2

        with pytest.raises(TypeError):
            struct_class.cast('')

    def test_iter_referenced_values(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        struct_type = ns.get_or_create_struct_type('test')
        struct_type.add_field(StructField(ns, 0, ns.Int.type, 'int_field'))
        struct_type.add_field(StructField(
            ns, 4, struct_type.get_pointer_type(), 'self_ptr'))

        struct_class = ns.get_class_for_type(struct_type)
        assert issubclass(struct_class, StructValue)

        arr_class = ns.array(struct_class, 1)
        arr = arr_class([struct_class(int_field=7)])
        struct = arr.values[0]
        assert isinstance(struct, StructValue)

        refs = list(struct.iter_referenced_values())
        assert len(refs) == 3
        assert arr in refs
        assert struct.int_field in refs
        assert struct.self_ptr in refs

    def test_unpack(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        struct_type = ns.get_or_create_struct_type('test')
        struct_type.add_field(StructField(ns, 0, ns.Int.type, 'int_field'))
        struct_type.add_field(StructField(
            ns, 4, struct_type.get_pointer_type(), 'self_ptr'))

        struct_class = ns.get_class_for_type(struct_type)
        assert issubclass(struct_class, StructValue)

        struct = struct_class.unpack_from_buffer(
            b'\x01\x00\x00\x00' b'\x02\x00\x00\x00')
        assert isinstance(struct.int_field, IntValue)
        assert int(struct.int_field) == 1
        assert isinstance(struct.self_ptr, PointerValue)
        assert int(struct.self_ptr) == 2

    def test_struct(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        struct_type = ns.get_or_create_struct_type('test')
        struct_type.add_field(StructField(ns, 0, ns.Int.type, 'int_field'))
        struct_type.add_field(StructField(
            ns, 4, struct_type.get_pointer_type(), 'self_ptr'))

        struct_class = ns.get_class_for_type(struct_type)
        assert issubclass(struct_class, StructValue)

        struct = struct_class(int_field=7)
        assert struct.type.name == 'test'
        assert isinstance(struct.int_field, IntValue)
        assert int(struct.int_field) == 7
        assert isinstance(struct.self_ptr, PointerValue)
        assert struct.self_ptr.pointed_address is None

        with pytest.raises(AttributeError):
            struct.not_a_real_field


class TestPointerValue:
    def test_pack(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        assert ns.VoidPointer.cast(0x04030201).pack() == (b'\x01\x02\x03\x04')
        assert ns.VoidPointer().pack() == (b'\x00\x00\x00\x00')

        ns = make_namespace(endian=Endianness.BIG, word_size=4)
        assert ns.VoidPointer.cast(0x04030201).pack() == (b'\x04\x03\x02\x01')

        ns = make_namespace(endian=Endianness.LITTLE, word_size=8)
        assert ns.VoidPointer.cast(0x0807060504030201).pack() == (
            b'\x01\x02\x03\x04\x05\x06\x07\x08')

        ns = make_namespace(endian=Endianness.BIG, word_size=8)
        assert ns.VoidPointer.cast(0x0807060504030201).pack() == (
            b'\x08\x07\x06\x05\x04\x03\x02\x01')

    def test_conversions(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        ptr = ns.VoidPointer(raw_value=0)
        assert int(ptr) == 0
        assert str(ptr) == "(void*)NULL"

        ptr = ns.VoidPointer(raw_value=1)
        assert int(ptr) == 1
        assert str(ptr) == "(void*)0x1"

        ptr = ns.VoidPointer(raw_value=1)
        assert int(ptr) == 1

        ptr = ns.VoidPointer()
        assert str(ptr) == "(void*)None"
        with pytest.raises(ValueError):
            int(ptr)

        ptr = ptr.ref()
        assert str(ptr) == "(void**)None"
        with pytest.raises(ValueError):
            int(ptr)

    def test_repr(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        ptr = ns.VoidPointer()
        assert repr(ptr) == "<PointerValue to void@None>"

        ptr = ns.VoidPointer(raw_value=0)
        assert repr(ptr) == "<PointerValue to void@0>"

    def test_get_address(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        ptr = ns.VoidPointer()
        assert ptr.pointed_address is None

        ptr_ptr = ptr.ref()
        assert ptr_ptr.pointed_address is None

    def test_copy(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        ptr = ns.VoidPointer.cast(0)
        copy = ptr.copy()
        assert ptr.pointed_address == 0
        assert copy.pointed_address == 0

        ptr.raw_value = 1
        assert ptr.pointed_address == 1
        assert copy.pointed_address == 0

        copy.raw_value = 2
        assert ptr.pointed_address == 1
        assert copy.pointed_address == 2

    def test_iter_referenced_values(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        ptr = ns.VoidPointer.cast(0)
        refs = list(ptr.iter_referenced_values())
        assert len(refs) == 0

        offset_ptr = ns.VoidPointer.cast(0, address_base=ptr, offset=4)
        refs = list(offset_ptr.iter_referenced_values())
        assert len(refs) == 1
        assert refs[0] is ptr

        ptr_ptr = ptr.ref()
        refs = list(ptr_ptr.iter_referenced_values())
        assert len(refs) == 1
        assert refs[0] is ptr

        offset_ptr_ptr = ptr.ref()
        offset_ptr_ptr.move(ptr, 4)
        refs = list(offset_ptr_ptr.iter_referenced_values())
        assert len(refs) == 2
        assert refs == [ptr, ptr]

    def test_cast(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        assert ns.VoidPointer.cast(0).pointed_address == 0

        char_pointer_class = ns.Char.get_pointer_class()
        test_pointer = char_pointer_class.cast('test')
        assert test_pointer.referenced_value is not None
        assert isinstance(test_pointer.referenced_value, ArrayValue)
        assert isinstance(test_pointer.referenced_value.type.member_type, IntType)
        assert test_pointer.referenced_value.type.member_type.size == 1
        assert str(test_pointer.referenced_value) == 'test'

        with pytest.raises(TypeError):
            ns.VoidPointer.cast(1.1)

    def test_unpack(self) -> None:
        ns = make_namespace(endian=Endianness.LITTLE, word_size=4)
        ptr = ns.VoidPointer.unpack_from_buffer(b'\x01\x02\x03\x04')
        assert ptr.pointed_address == 0x04030201

        ns = make_namespace(endian=Endianness.BIG, word_size=4)
        ptr = ns.VoidPointer.unpack_from_buffer(b'\x01\x02\x03\x04')
        assert ptr.pointed_address == 0x01020304

        ns = make_namespace(endian=Endianness.LITTLE, word_size=8)
        ptr = ns.VoidPointer.unpack_from_buffer(b'\x01\x02\x03\x04\x05\x06\x07\x08')
        assert ptr.pointed_address == 0x0807060504030201

        ns = make_namespace(endian=Endianness.BIG, word_size=8)
        ptr = ns.VoidPointer.unpack_from_buffer(b'\x01\x02\x03\x04\x05\x06\x07\x08')
        assert ptr.pointed_address == 0x0102030405060708

    def test_pointer(self) -> None:
        for ns in make_namespaces():
            null_pointer = ns.VoidPointer(raw_value=0)
            assert null_pointer.type.size == ns.arch.pointer_size
            assert null_pointer.pointed_address == 0
            assert null_pointer.address == None
            with pytest.raises(ValueError):
                ns.VoidPointer(referenced_value=null_pointer, raw_value=0)


class TestArrayValue:
    def test_pack(self) -> None:
        ns = amd64_namespace()
        char_arr_class = ns.array(ns.Char, 3)
        char_arr = char_arr_class.cast("hi")
        assert char_arr.pack() == b'hi\x00'

        arr_class = ns.array(ns.Int, 3)
        int_arr = arr_class.cast([0, 1, 2])
        assert int_arr.pack() == (
            b'\x00\x00\x00\x00'
            b'\x01\x00\x00\x00'
            b'\x02\x00\x00\x00'
        )

    def test_conversions(self) -> None:
        ns = amd64_namespace()
        char_arr_class = ns.array(ns.Char, 3)
        char_arr = char_arr_class.cast("hi")

        assert bytes(char_arr) == b'hi'
        assert str(char_arr) == 'hi'
        assert repr(char_arr).replace('"', "'") in "(char[3])b'hi'"

        arr_class = ns.array(ns.Int, 3)
        int_arr = arr_class.cast([0, 1, 2])
        assert repr(int_arr) == '{0, 1, 2}'

        with pytest.raises(TypeError):
            bytes(int_arr)

    def test_cast(self) -> None:
        ns = amd64_namespace()
        arr_class = ns.array(ns.Int, 3)
        arr = arr_class.cast([0, 1, 2])
        assert len(arr) == 3
        assert [int(v) for v in arr.values] == [0, 1, 2]  # type: ignore

        arr = arr_class.cast([1])
        assert len(arr) == 3
        assert int(arr.values[0]) == 1  # type: ignore
        with pytest.raises(ValueError):
            int(arr.values[1])  # type: ignore

        with pytest.raises(TypeError):
            arr_class.cast(0)
        with pytest.raises(TypeError):
            arr_class.cast([0, 1, 2, 3])

    def test_iter_referenced_values(self) -> None:
        ns = amd64_namespace()
        arr_class = ns.array(ns.Int, 3)
        buf = b''.join(ns.Int(i).pack() for i in range(3))
        arr = arr_class.unpack_from_buffer(buf)
        refs = list(arr.iter_referenced_values())
        assert len(refs) == 3
        assert all(isinstance(ref, IntValue) for ref in refs)
        assert sorted(int(ref) for ref in refs) == [0, 1, 2]  # type: ignore

        arr2_class = ns.array(arr_class, 1)
        arr2 = arr2_class.unpack_from_buffer(buf)
        arr = arr2.values[0]

        refs = list(arr.iter_referenced_values())
        assert len(refs) == 4
        int_refs = []
        for ref in refs:
            if isinstance(ref, IntValue):
                int_refs.append(int(ref))
            else:
                assert ref is arr2
        assert sorted(int_refs) == [0, 1, 2]

    def test_array_unpack(self) -> None:
        ns = amd64_namespace()
        arr_class = ns.array(ns.Int, 3)
        buf = b''.join(ns.Int(i).pack() for i in range(3))
        arr = arr_class.unpack_from_buffer(buf)
        assert len(arr) == 3

    def test_array(self) -> None:
        ns = amd64_namespace()
        arr_class = ns.array(ns.Int, 3)
        assert arr_class.type.count == 3

        arr = arr_class()
        assert len(arr.values) == 3

        arr = arr_class.cast([1, 2, 3])
        assert len(arr.values) == 3

        with pytest.raises(ValueError):
            arr_class(values=[])


class TestValue:
    def test_ref(self) -> None:
        num1 = make_namespace(endian=Endianness.LITTLE, word_size=8).Int(
            0, offset=4
        )
        ptr = num1.ref()
        assert isinstance(ptr, PointerValue)
        assert ptr.referenced_value is num1
        assert ptr.pointed_address == 4

    def test_address_and_move(self) -> None:
        num1 = make_namespace(endian=Endianness.LITTLE, word_size=8).Int(
            0, offset=0
        )
        num2 = make_namespace(endian=Endianness.LITTLE, word_size=8).Int(
            1, offset=4
        )
        num3 = make_namespace(endian=Endianness.LITTLE, word_size=8).Int(
            2, address_base=num2, offset=4
        )
        assert num1.address == 0
        assert num2.address == 4
        assert num3.address == 8

        num2.move(None, 8)
        assert num1.address == 0
        assert num2.address == 8
        assert num3.address == 12

        num2.move(None, None)
        assert num1.address == 0
        assert num2.address is None
        assert num3.address is None

    def test_abstract_methods(self) -> None:
        # Instantiate a subclass of Value to invoke abstract Value methods on
        val_obj = make_namespace(endian=Endianness.LITTLE, word_size=8).Int(0)
        with pytest.raises(NotImplementedError):
            Value.unpack_from_buffer(b"")
        with pytest.raises(NotImplementedError):
            Value.cast(0)
        with pytest.raises(NotImplementedError):
            Value.copy(val_obj)
        with pytest.raises(NotImplementedError):
            Value.pack(val_obj)
        with pytest.raises(NotImplementedError):
            Value.iter_referenced_values(val_obj)


class TestInt:
    def test_str_repr(self) -> None:
        for ns in make_namespaces():
            assert str(ns.Int()) == "<uninitialized>"
            assert str(ns.Int(1)) == "1"
            assert str(ns.Int(-1)) == "-1"

            assert repr(ns.Int()) == "<int <uninitialized>>"
            assert repr(ns.Int(1)) == "<int 1>"
            assert repr(ns.Int(-1)) == "<int -1>"

    def test_conversions(self) -> None:
        for ns in make_namespaces():
            a = ns.Int(3)
            assert float(a) == 3.0
            assert int(a) == 3

            with pytest.raises(ValueError):
                float(ns.Int())
            with pytest.raises(ValueError):
                int(ns.Int())

    def test_copy(self) -> None:
        for ns in make_namespaces():
            a = ns.Int(0)
            assert a.value == 0

            b = a.copy()
            assert isinstance(b, IntValue)
            assert a is not b
            assert b.value == 0

            a.value = 1
            assert a.value == 1
            assert b.value == 0

            b.value = 2
            assert a.value == 1
            assert b.value == 2

    def test_iter_referenced_values(self) -> None:
        for ns in make_namespaces():
            a = ns.Int(0)
            b = ns.Int(0, address_base=a, offset=4)
            assert len(list(a.iter_referenced_values())) == 0
            assert len(list(b.iter_referenced_values())) == 1
            assert list(b.iter_referenced_values())[0] is a

    def test_no_size(self) -> None:
        for ns in make_namespaces():
            partial_type = ns.get_class_for_type(
                IntType(ns, "partial_int_t", None, False)
            )
            with pytest.raises(ValueError):
                partial_type.cast(0)

    def test_cast_int(self) -> None:
        for ns in make_namespaces():
            assert ns.Int.cast(0).value == 0
            assert ns.Int.cast(1).value == 1
            assert ns.Int.cast("a").value == ord("a")
            assert ns.Int.cast(b"a").value == ord(b"a")
            assert ns.Int.cast(1.1).value == 1
            assert ns.Int.cast(1.5).value == 1
            assert ns.Int.cast(1.9).value == 1
            assert ns.Int.cast(-2.8).value == -2
            assert ns.Int.cast(0x7FFFFFFF).value == 0x7FFFFFFF
            assert ns.Int.cast(0xFFFFFFFF).value == -1
            assert ns.Int.cast(0xFFFFFFFE).value == -2
            assert ns.Int.cast(-1).value == -1
            assert ns.Int.cast(-2).value == -2
            with pytest.raises(TypeError):
                ns.Int.cast([])
            with pytest.raises(TypeError):
                ns.Int.cast("string")

    def test_cast_uint(self) -> None:
        for ns in make_namespaces():
            assert ns.UnsignedInt.cast(0).value == 0
            assert ns.UnsignedInt.cast(1).value == 1
            assert ns.UnsignedInt.cast("a").value == ord("a")
            assert ns.UnsignedInt.cast(b"a").value == ord(b"a")
            assert ns.UnsignedInt.cast(1.1).value == 1
            assert ns.UnsignedInt.cast(1.5).value == 1
            assert ns.UnsignedInt.cast(1.9).value == 1
            assert ns.UnsignedInt.cast(-2.8).value == 0xFFFFFFFE
            assert ns.UnsignedInt.cast(0x7FFFFFFF).value == 0x7FFFFFFF
            assert ns.UnsignedInt.cast(0xFFFFFFFF).value == 0xFFFFFFFF
            assert ns.UnsignedInt.cast(0xFFFFFFFE).value == 0xFFFFFFFE
            assert ns.UnsignedInt.cast(-1).value == 0xFFFFFFFF
            assert ns.UnsignedInt.cast(-2).value == 0xFFFFFFFE

    def test_underflow_overflow(self) -> None:
        unsigned_error = ".*cannot be represented by unsigned int of size.*"
        signed_error = ".*cannot be represented by signed int of size.*"
        for ns in make_namespaces():
            with pytest.raises(ValueError, match=signed_error):
                ns.Int(-2147483649)
            with pytest.raises(ValueError, match=signed_error):
                ns.Int(0x80000000)
            with pytest.raises(ValueError, match=signed_error):
                ns.Int(0x100000000)
            with pytest.raises(ValueError, match=unsigned_error):
                ns.UnsignedInt(-1)
            with pytest.raises(ValueError, match=unsigned_error):
                ns.UnsignedInt(0x100000000)
            with pytest.raises(ValueError, match=signed_error):
                ns.Char(128)
            with pytest.raises(ValueError, match=unsigned_error):
                ns.UnsignedChar(-1)

    def test_pack_integral_overflow(self) -> None:
        with pytest.raises(ValueError):
            _pack_integral_type(-1, 1, False, Endianness.LITTLE)

    def test_pack_int(self) -> None:
        def ntoh(b: bytes) -> bytes:
            if ns.arch.endianness == Endianness.LITTLE:
                return b[::-1]
            return b

        for ns in make_namespaces():
            assert ns.Int().pack() == ntoh(b"\x00\x00\x00\x00")
            assert ns.Int(1).pack() == ntoh(b"\x00\x00\x00\x01")
            assert ns.Int(2).pack() == ntoh(b"\x00\x00\x00\x02")
            assert ns.Int(0x7FFFFFFF).pack() == ntoh(b"\x7f\xff\xff\xff")
            assert ns.Int(-1).pack() == ntoh(b"\xff\xff\xff\xff")
            assert ns.Int(-2).pack() == ntoh(b"\xff\xff\xff\xfe")

            assert ns.UnsignedInt().pack() == ntoh(b"\x00\x00\x00\x00")
            assert ns.UnsignedInt(0).pack() == ntoh(b"\x00\x00\x00\x00")
            assert ns.UnsignedInt(1).pack() == ntoh(b"\x00\x00\x00\x01")
            assert ns.UnsignedInt(0xFFFFFFFF).pack() == ntoh(
                b"\xff\xff\xff\xff"
            )

    def test_unpack_int(self) -> None:
        for ns in make_namespaces():

            def ntoh(b: bytes) -> bytes:
                if ns.arch.endianness == Endianness.LITTLE:
                    return b[::-1]
                return b

            assert (
                ns.Int.unpack_from_buffer(ntoh(b"\x00\x00\x00\x00")).value == 0
            )
            assert (
                ns.Int.unpack_from_buffer(ntoh(b"\x00\x00\x00\x01")).value == 1
            )
            assert (
                ns.Int.unpack_from_buffer(ntoh(b"\x7f\xff\xff\xff")).value
                == 0x7FFFFFFF
            )
            assert (
                ns.Int.unpack_from_buffer(ntoh(b"\xff\xff\xff\xff")).value
                == -1
            )
            assert (
                ns.Int.unpack_from_buffer(ntoh(b"\xff\xff\xff\xfe")).value
                == -2
            )
            with pytest.raises(ValueError):
                ns.Int.unpack_from_buffer(b"")
            with pytest.raises(ValueError):
                ns.Int.unpack_from_buffer(b"123")
            with pytest.raises(ValueError):
                ns.Int.unpack_from_buffer(b"12345")

    def test_unpack_unsigned(self) -> None:
        for ns in make_namespaces():

            def ntoh(b: bytes) -> bytes:
                if ns.arch.endianness == Endianness.LITTLE:
                    return b[::-1]
                return b

            assert (
                ns.UnsignedInt.unpack_from_buffer(
                    ntoh(b"\x00\x00\x00\x00")
                ).value
                == 0
            )
            assert (
                ns.UnsignedInt.unpack_from_buffer(
                    ntoh(b"\x00\x00\x00\x01")
                ).value
                == 1
            )
            assert (
                ns.UnsignedInt.unpack_from_buffer(
                    ntoh(b"\x7f\xff\xff\xff")
                ).value
                == 0x7FFFFFFF
            )
            assert (
                ns.UnsignedInt.unpack_from_buffer(
                    ntoh(b"\xff\xff\xff\xff")
                ).value
                == 0xFFFFFFFF
            )
            assert (
                ns.UnsignedInt.unpack_from_buffer(
                    ntoh(b"\xff\xff\xff\xfe")
                ).value
                == 0xFFFFFFFE
            )
            with pytest.raises(ValueError):
                ns.UnsignedInt.unpack_from_buffer(b"")
            with pytest.raises(ValueError):
                ns.UnsignedInt.unpack_from_buffer(b"123")
            with pytest.raises(ValueError):
                ns.UnsignedInt.unpack_from_buffer(b"12345")



from moria.util import hexdump, SortedList
import pytest


def test_hexdump(capfd: pytest.CaptureFixture):
    hexdump(b'hello\x00\x01\xff', start_address=0x10)
    out, err = capfd.readouterr()
    assert err == ''
    assert out == "0000000000000010  68 65 6c 6c 6f 00 01 ff                           |hello...|\n"


def test_sorted_list():
    def comp_ints(x: int, y: int) -> int:
        return x - y
    nums = SortedList(key=comp_ints)
    nums_to_insert = [5, 4, 2, 5123123, -5123, -51, 0, 4, 1]
    for n in nums_to_insert:
        nums.insert(n)
    assert list(nums) == sorted(nums_to_insert)



import pytest
from tests.helpers import amd64_namespace
from moria.basetypes import Type
from typing import Any


class TestNamespace:
    def test_no_superclass(self) -> None:
        ns = amd64_namespace()

        with pytest.raises(TypeError):
            ns._get_superclass_for_type(None) # type: ignore

    def test_conflicting_struct_name(self) -> None:
        ns = amd64_namespace()
        ns.get_or_create_struct_type('__init__')
        with pytest.raises(ValueError):
            ns.initialize_struct_classes()

    def test_struct_name(self) -> None:
        ns = amd64_namespace()
        assert ns._format_struct_name('test') == 'test'
        with pytest.raises(ValueError):
            ns._format_struct_name('!')

    def test_pack_cyclic(self) -> None:
        ns = amd64_namespace()
        i = ns.Int()
        j = ns.Int()

        # Cyclic address dependencies are not supported, even when consistent.
        i.address_base = j
        i.offset = -4
        j.address_base = i
        j.offset = 4

        with pytest.raises(ValueError):
            ns.pack_values(0x0, 0x100, [i])

    def test_pack_implicit(self) -> None:
        ns = amd64_namespace()
        char_arr = ns.array(ns.Char, 4)
        test_char_arr = char_arr.cast("test")
        test_ptr = test_char_arr.values[0].ref()
        packed_vals = ns.pack_values(0x0, 12, [test_ptr])

        # The allocator may choose any order to pack the values, so
        # there are two possible valid packings that are OK to return.
        assert packed_vals in (
            b'test\x00\x00\x00\x00\x00\x00\x00\x00',
            b'\x08\x00\x00\x00\x00\x00\x00\x00test',
        )

    def test_pack(self) -> None:
        ns = amd64_namespace()
        assert ns.pack_values(0x0, 0x1000, []) == b''

        i = ns.UInt32(0xdeadbeef)
        assert ns.pack_values(0x0, 0x1000, [i]) == b'\xef\xbe\xad\xde'

        i_ptr = i.ref()
        packed_vals = ns.pack_values(0x0, 12, [i, i_ptr])

        # The allocator may choose any order to pack the values, so
        # there are two possible valid packings that are OK to return.
        assert packed_vals in (
            b'\xef\xbe\xad\xde\x00\x00\x00\x00\x00\x00\x00\x00',
            b'\x08\x00\x00\x00\x00\x00\x00\x00\xef\xbe\xad\xde',
        )

        # This is invalid because there is not enough space
        with pytest.raises(ValueError):
            ns.pack_values(0x0, 11, [i, i_ptr])

        # This is invalid because the objects already have fixed
        # addresses incompatible with the given address range.
        with pytest.raises(ValueError):
            ns.pack_values(0x10, 12, [i, i_ptr])



from __future__ import annotations
from typing import (
    Iterable,
    MutableMapping,
    List,
    Set,
)

# we already have a "Type" class, so we need to avoid a name conflict here
from typing import Type as PyType
import re
from moria.arch import Arch
from moria.pack import HeapPacker

import moria.basetypes as basetypes
from moria.values import ArrayValue, PointerValue, Value, StructValue, IntValue


class Namespace:
    arch: Arch
    struct_types: MutableMapping[str, PyType[StructValue]]

    Char: PyType[IntValue]
    UnsignedChar: PyType[IntValue]
    Short: PyType[IntValue]
    UnsignedShort: PyType[IntValue]
    Int: PyType[IntValue]
    UnsignedInt: PyType[IntValue]
    Long: PyType[IntValue]
    UnsignedLong: PyType[IntValue]
    LongLong: PyType[IntValue]
    UnsignedLongLong: PyType[IntValue]

    Int8: PyType[IntValue]
    UInt8: PyType[IntValue]
    Int16: PyType[IntValue]
    UInt16: PyType[IntValue]
    Int32: PyType[IntValue]
    UInt32: PyType[IntValue]
    Int64: PyType[IntValue]
    UInt64: PyType[IntValue]

    VoidPointer: PyType[PointerValue]

    _type_classes: MutableMapping[str, PyType[Value]]

    def __init__(self, arch: Arch) -> None:
        self.arch = arch
        self.structs: MutableMapping[str, basetypes.StructType] = {}
        self.struct_types = {}
        self._type_classes = {}
        self._create_default_types()

    def _create_typed_integer_class(
        self, name: str, size: int, signed: bool
    ) -> PyType[IntValue]:
        int_class = self.get_class_for_type(
            basetypes.IntType(self, name, size, _signed=signed)
        )
        assert issubclass(int_class, IntValue)
        return int_class

    def _create_default_types(self) -> None:
        self.Char = self._create_typed_integer_class(
            "char", self.arch.char_size, True
        )
        self.UnsignedChar = self._create_typed_integer_class(
            "unsigned char", self.arch.char_size, False
        )
        self.Short = self._create_typed_integer_class(
            "short", self.arch.short_size, True
        )
        self.UnsignedShort = self._create_typed_integer_class(
            "unsigned short", self.arch.short_size, False
        )
        self.Int = self._create_typed_integer_class(
            "int", self.arch.int_size, True
        )
        self.UnsignedInt = self._create_typed_integer_class(
            "unsigned int", self.arch.int_size, False
        )
        self.Long = self._create_typed_integer_class(
            "long", self.arch.long_size, True
        )
        self.UnsignedLong = self._create_typed_integer_class(
            "unsigned long", self.arch.long_size, False
        )
        self.LongLong = self._create_typed_integer_class(
            "long long", self.arch.long_long_size, True
        )
        self.UnsignedLongLong = self._create_typed_integer_class(
            "unsigned long long", self.arch.long_long_size, False
        )

        self.Int8 = self._create_typed_integer_class("int8_t", 1, True)
        self.UInt8 = self._create_typed_integer_class("uint8_t", 1, False)
        self.Int16 = self._create_typed_integer_class("int16_t", 2, True)
        self.UInt16 = self._create_typed_integer_class("uint16_t", 2, False)
        self.Int32 = self._create_typed_integer_class("int32_t", 4, True)
        self.UInt32 = self._create_typed_integer_class("uint32_t", 4, False)
        self.Int64 = self._create_typed_integer_class("int64_t", 8, True)
        self.UInt64 = self._create_typed_integer_class("uint64_t", 8, False)

        void_pointer_class = self.get_class_for_type(
            basetypes.IntType(self, 'void', None, False).get_pointer_type())
        assert(issubclass(void_pointer_class, PointerValue))
        self.VoidPointer = void_pointer_class

    def array(self, t: PyType[Value], count: int) -> PyType[ArrayValue]:
        array_type = basetypes.ArrayType(self, t.type, count)
        array_class = self.get_class_for_type(array_type)
        assert issubclass(array_class, ArrayValue)
        return array_class

    def pack_values(
        self, base_address: int, max_size: int, values: Iterable[Value]
    ) -> bytes:
        all_values: Set[Value] = set()
        to_traverse: List[Value] = list(values)
        while len(to_traverse) > 0:
            val = to_traverse.pop()
            all_values.add(val)
            for child in val.iter_referenced_values():
                if child not in all_values:
                    to_traverse.append(child)

        free_values: Set[Value] = set()
        fixed_values: Set[Value] = set()
        for val in all_values:
            parent = val
            path_to_parent: List[Value] = [val]
            while parent.address_base is not None:
                parent = parent.address_base
                if parent in path_to_parent:
                    path_to_parent.append(parent)
                    path_str = ' -> '.join(repr(v) for v in path_to_parent)
                    raise ValueError(f'Cyclic address dependency: {path_str}')
                path_to_parent.append(parent)
            if parent.offset is None:
                free_values.add(parent)
            else:
                fixed_values.add(parent)

        packer = HeapPacker(base_address, max_size)
        for val in fixed_values:
            assert val.type.size is not None
            assert val.address is not None
            packer.alloc(val.type.size, val.address)

        for val in free_values:
            assert val.type.size is not None
            val.offset = packer.alloc(val.type.size)

        parts: List[bytes] = []
        vals_to_pack: List[Value] = list(free_values) + list(fixed_values)
        vals_to_pack.sort(key=lambda val: val.address or 0)
        for i, val in enumerate(vals_to_pack):
            assert val.address is not None
            assert val.type.size is not None
            part_end = val.address + val.type.size
            if i < len(vals_to_pack) - 1:
                next_start = vals_to_pack[i+1].address
                assert next_start is not None
                pad = next_start - part_end
            else:
                pad = 0
            parts.append(val.pack() + pad*b'\x00')

        return b"".join(parts)

    def get_or_create_struct_type(self, name: str) -> basetypes.StructType:
        struct = self.structs.get(name)
        if struct is not None:
            return struct
        struct = basetypes.StructType(self, name)
        self.structs[name] = struct
        return struct

    def _format_struct_name(self, name: str) -> str:
        if re.fullmatch("[a-zA-Z_][a-zA-Z_0-9]*", name) is None:
            raise ValueError(f"Invalid struct name {name}")
        return name

    def __getattr__(self, name: str) -> PyType[StructValue]:
        if name in self.struct_types:
            return self.struct_types[name]
        raise AttributeError

    def initialize_struct_classes(self) -> None:
        for struct_type in self.structs.values():
            struct_name = self._format_struct_name(struct_type.name)
            struct_class = self.get_class_for_type(struct_type)
            assert issubclass(struct_class, StructValue)
            if hasattr(self, struct_name):
                raise ValueError(
                    f"Struct name {struct_name} conflicts "
                    "with existing namespace field"
                )
            self.struct_types[struct_name] = struct_class

    def _get_superclass_for_type(self, t: basetypes.Type) -> PyType[Value]:
        if isinstance(t, basetypes.IntType):
            return IntValue
        elif isinstance(t, basetypes.PointerType):
            return PointerValue
        elif isinstance(t, basetypes.ArrayType):
            return ArrayValue
        elif isinstance(t, basetypes.StructType):
            return StructValue
        raise TypeError("No superclass for type {t}!")

    def get_class_for_type(self, t: basetypes.Type) -> PyType[Value]:
        if t.name in self._type_classes:
            return self._type_classes[t.name]
        else:
            superclass = self._get_superclass_for_type(t)

            class type_class(superclass):
                type = t

            self._type_classes[t.name] = type_class
            return type_class

    def print_structs(self):
        for i, struct in enumerate(self.structs.values()):
            if i > 0:
                print()
            print(struct.pretty_string())



from moria.namespace import Namespace
from moria.arch import Arch, Endianness
from typing import Iterable

def make_namespaces() -> Iterable[Namespace]:
    endians = (Endianness.LITTLE, Endianness.BIG)
    word_sizes = (4, 8)
    for endian in endians:
        for word_size in word_sizes:
            yield make_namespace(endian, word_size)


def amd64_namespace() -> Namespace:
    return make_namespace(Endianness.LITTLE, 8)


def make_namespace(endian: Endianness, word_size: int) -> Namespace:
    return Namespace(arch=Arch(endianness=endian, word_size=word_size))


from enum import Enum


class Endianness(Enum):
    LITTLE = 0
    BIG = 1


class Arch:
    endianness: Endianness
    char_size: int
    short_size: int
    int_size: int
    long_size: int
    long_long_size: int
    pointer_size: int

    def __init__(self, endianness: Endianness, word_size: int):
        self.endianness = endianness
        assert word_size in (4, 8), "Invalid word size"

        self.char_size = 1
        self.short_size = 2
        self.long_long_size = 8
        self.pointer_size = word_size

        if word_size == 4:
            self.int_size = 4
            self.long_size = 4
        elif word_size == 8:
            self.int_size = 4
            self.long_size = 8
